@@ -5,6 +5,8 @@ begin
 rescue LoadError
 end
 
+module DRbTests
+
 if Object.const_defined?("UNIXServer")
 
 
@@ -20,8 +22,13 @@ end
 class TestDRbUNIXCore < Test::Unit::TestCase
   include DRbCore
   def setup
-    @ext = DRbUNIXService.ext_service('ut_drb_drbunix.rb')
-    @there = @ext.front
+    setup_service 'ut_drb_drbunix.rb'
+    super
+  end
+
+  def teardown
+    super
+    DRbService.finish
   end
 
   def test_02_unknown
@@ -37,10 +44,16 @@ end
 class TestDRbUNIXAry < Test::Unit::TestCase
   include DRbAry
   def setup
-    @ext = DRbUNIXService.ext_service('ut_array_drbunix.rb')
-    @there = @ext.front
+    setup_service 'ut_array_drbunix.rb'
+    super
+  end
+  def teardown
+    super
+    DRbService.finish
   end
 end
 
 
+end
+
 end
