@@ -37,7 +37,6 @@ class MockClock
   def _forward(n=nil)
     now ,= @ts.take([nil, :now])
     @now = now + n
-    n = @reso if n.nil?
     @ts.write([@now, :now])
   end
 
@@ -449,7 +448,12 @@ class TupleSpaceTest < Test::Unit::TestCase
   end
   def teardown
     # implementation-dependent
-    @ts.instance_eval{@keeper.kill if @keeper}
+    @ts.instance_eval{
+      if th = @keeper
+        th.kill
+        th.join
+      end
+    }
   end
 end
 
@@ -463,7 +467,12 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
   end
   def teardown
     # implementation-dependent
-    @ts_base.instance_eval{@keeper.kill if @keeper}
+    @ts_base.instance_eval{
+      if th = @keeper
+        th.kill
+        th.join
+      end
+    }
   end
 
   def test_remote_array_and_hash
@@ -479,8 +488,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
   end
 
   def test_take_bug_8215
-    require_relative '../ruby/envutil'
-    service = DRb.start_service("druby://localhost:0", @ts_base)
+    service = DRb.start_service(nil, @ts_base)
 
     uri = service.uri
 
@@ -488,7 +496,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
 
     take = spawn(*args, <<-'end;', uri)
       uri = ARGV[0]
-      DRb.start_service("druby://localhost:0")
+      DRb.start_service
       ro = DRbObject.new_with_uri(uri)
       ts = Rinda::TupleSpaceProxy.new(ro)
       th = Thread.new do
@@ -504,7 +512,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
 
     write = spawn(*args, <<-'end;', uri)
       uri = ARGV[0]
-      DRb.start_service("druby://localhost:0")
+      DRb.start_service
       ro = DRbObject.new_with_uri(uri)
       ts = Rinda::TupleSpaceProxy.new(ro)
       ts.write([:test_take, 42])
@@ -515,6 +523,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
     assert_equal([[:test_take, 42]], @ts_base.read_all([:test_take, nil]),
                  '[bug:8215] tuple lost')
   ensure
+    service.stop_service if service
     signal = /mswin|mingw/ =~ RUBY_PLATFORM ? "KILL" : "TERM"
     Process.kill(signal, write) if write && status.nil?
     Process.kill(signal, take)  if take
@@ -522,7 +531,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
     Process.wait(take)  if take
   end
 
-  @server = DRb.primary_server || DRb.start_service("druby://localhost:0")
+  @server = DRb.primary_server || DRb.start_service
 end
 
 module RingIPv6
@@ -554,7 +563,12 @@ class TestRingServer < Test::Unit::TestCase
   end
   def teardown
     # implementation-dependent
-    @ts.instance_eval{@keeper.kill if @keeper}
+    @ts.instance_eval{
+      if th = @keeper
+        th.kill
+        th.join
+      end
+    }
     @rs.shutdown
   end
 
@@ -641,6 +655,7 @@ class TestRingServer < Test::Unit::TestCase
   end
 
   def test_ring_server_ipv4_multicast
+    @rs.shutdown
     @rs = Rinda::RingServer.new(@ts, [['239.0.0.1', '0.0.0.0']], @port)
     v4mc = @rs.instance_variable_get('@sockets').first
 
@@ -658,6 +673,7 @@ class TestRingServer < Test::Unit::TestCase
     skip 'IPv6 not available' unless
       Socket.ip_address_list.any? { |addrinfo| addrinfo.ipv6? }
 
+    @rs.shutdown
     begin
       @rs = Rinda::RingServer.new(@ts, [['ff02::1', '::1', 0]], @port)
     rescue Errno::EADDRNOTAVAIL
@@ -709,7 +725,10 @@ class TestRingServer < Test::Unit::TestCase
     end
     raise Timeout::Error, "timeout", bt
   ensure
-    th.kill if th
+    if th
+      th.kill
+      th.join
+    end
     Thread.abort_on_exception = aoe
   end
 
@@ -735,6 +754,8 @@ class TestRingFinger < Test::Unit::TestCase
     v4 = @rf.make_socket('127.0.0.1')
 
     assert(v4.getsockopt(:SOL_SOCKET, :SO_BROADCAST).bool)
+  ensure
+    v4.close if v4
   end
 
   def test_make_socket_ipv4_multicast
@@ -742,6 +763,8 @@ class TestRingFinger < Test::Unit::TestCase
 
     assert_equal(1, v4mc.getsockopt(:IPPROTO_IP, :IP_MULTICAST_LOOP).ipv4_multicast_loop)
     assert_equal(1, v4mc.getsockopt(:IPPROTO_IP, :IP_MULTICAST_TTL).ipv4_multicast_ttl)
+  ensure
+    v4mc.close if v4mc
   end
 
   def test_make_socket_ipv6_multicast
@@ -755,12 +778,16 @@ class TestRingFinger < Test::Unit::TestCase
 
     assert_equal(1, v6mc.getsockopt(:IPPROTO_IPV6, :IPV6_MULTICAST_LOOP).int)
     assert_equal(1, v6mc.getsockopt(:IPPROTO_IPV6, :IPV6_MULTICAST_HOPS).int)
+  ensure
+    v6mc.close if v6mc
   end
 
   def test_make_socket_ipv4_multicast_hops
     @rf.multicast_hops = 2
     v4mc = @rf.make_socket('239.0.0.1')
     assert_equal(2, v4mc.getsockopt(:IPPROTO_IP, :IP_MULTICAST_TTL).ipv4_multicast_ttl)
+  ensure
+    v4mc.close if v4mc
   end
 
   def test_make_socket_ipv6_multicast_hops
@@ -773,6 +800,8 @@ class TestRingFinger < Test::Unit::TestCase
       v6mc = @rf.make_socket("ff02::1%#{ifaddr.name}")
     end
     assert_equal(2, v6mc.getsockopt(:IPPROTO_IPV6, :IPV6_MULTICAST_HOPS).int)
+  ensure
+    v6mc.close if v6mc
   end
 
 end
