@@ -1,9 +1,9 @@
-require "rss-testcase"
+require_relative "rss-testcase"
 
 module RSS
   class TestVersion < TestCase
     def test_version
-      assert_equal("0.2.5", ::RSS::VERSION)
+      assert_equal("0.2.7", ::RSS::VERSION)
     end
   end
 end
